import os
import sys
import json
import numpy as np
from tqdm import tqdm
from transformers import BartTokenizer
import torch
import torch.nn as nn
import random
import ujson

import pytorch_lightning as pl
from torch.utils.data import DataLoader
from datasets import load_dataset
from bioel.utils.bigbio_utils import (
    dataset_to_documents,
    dataset_to_df,
    resolve_abbreviation,
    load_dataset_df,
)
from bioel.utils.dataset_consts import (
    CUIS_TO_REMAP,
    CUIS_TO_EXCLUDE,
    VALIDATION_DOCUMENT_IDS,
)
from bioel.models.biogenel.data_processing import data_preprocess
from bioel.models.biogenel.trie.create_trie_and_target_kb import create_trie
from bioel.ontology import BiomedicalOntology


class MedMentionsDataset(torch.utils.data.Dataset):
    def __init__(self, encodings, labels, test_set=False):
        self.encodings = encodings
        self.labels = labels
        self.test_set = test_set

    def __getitem__(self, idx):
        item = {key: torch.tensor(val[idx]) for key, val in self.encodings.items()}
        item["label_ids"] = torch.tensor(self.labels["labels"][idx])
        item["decoder_input_ids"] = torch.tensor(self.labels["decoder_input_ids"][idx])
        item["decoder_attention_mask"] = torch.tensor(
            self.labels["attention_mask"][idx]
        )
        # the decoder atten mask has the same length as label of decoder input
        if self.test_set:
            item["decoder_input_ids_test"] = torch.tensor(
                self.labels["decoder_input_ids_test"][idx]
            )
            item["decoder_attention_mask_test"] = torch.tensor(
                self.labels["attention_mask_test"][idx]
            )

        # if self.labels['unlikelihood_tokens']:
        #     item['unlikelihood_mask'] = torch.zeros()
        return item

    def __len__(self):
        return len(self.labels["labels"])


def read_file_bylines(file):
    with open(file, "r") as f:
        lines = f.readlines()
    output = []
    for item in lines[:]:
        output.append(item.strip("\n"))
    return output


def encode_data_to_json(dataset_path, tokenizer):
    fi = [dataset_path + f for f in [".source", ".target"]]
    with open(fi[1] + ".token.json", "w") as f:
        for x in tqdm(read_file_bylines(fi[1])):
            xi = ujson.loads(x)
            lines = [
                list(tokenizer(" " + xi[0])["input_ids"])[1:-1],
                list(tokenizer(" " + xi[1])["input_ids"])[1:-1],
            ]
            lines = ujson.dumps(lines, ensure_ascii=False)
            f.write(lines + "\n")

    with open(fi[0] + ".token.json", "w") as f:
        for x in tqdm(read_file_bylines(fi[0])):
            # print(x)

            ix = ujson.loads(x)[0]
            line = list(tokenizer(" " + ix)["input_ids"])[1:-1]
            line = ujson.dumps([line], ensure_ascii=False)
            f.write(line + "\n")


def read_ids_from_json(path, prefix_mention_is=False):
    files = [path + f for f in [".source", ".target"]]
    tokens_x = {"input_ids": [], "attention_mask": []}
    tokens_y = {
        "labels": [],
        "attention_mask": [],
        "decoder_input_ids": [],
        "decoder_input_ids_test": [],
        "attention_mask_test": [],
        "unlikelihood_tokens": [],
    }
    max_len_x = 0
    max_len_y = 0

    for x, y in tqdm(
        zip(
            read_file_bylines(files[0] + ".token.json"),
            read_file_bylines(files[1] + ".token.json"),
        )
    ):
        x = ujson.loads(x)
        y = ujson.loads(y)

        prefix = y[0]
        label = y[1]
        y = sum(y, [])

        max_len_x = np.max([max_len_x, len(sum(x, [])) + 3])
        max_len_y = np.max([max_len_y, len(y) + 2])

        tokens_x["input_ids"].append([0] + x[0] + [2])
        tokens_x["attention_mask"].append(list(np.ones_like(tokens_x["input_ids"][-1])))

        if prefix_mention_is:
            tokens_y["decoder_input_ids"].append([2] + y)  # decoder input
            labs_prefix = [-100] * len(prefix) + label + [2]
            tokens_y["labels"].append(labs_prefix)
            assert len(labs_prefix) == len(y) + 1
        else:
            tokens_y["decoder_input_ids"].append([2] + label)  # decoder input
            tokens_y["labels"].append(label + [2])  # labels
        tokens_y["attention_mask"].append(
            list(np.ones_like(tokens_y["decoder_input_ids"][-1]))
        )

        if "test" in path:
            if prefix_mention_is:
                tokens_y["decoder_input_ids_test"].append(prefix)
            else:
                tokens_y["decoder_input_ids_test"].append([2])
            tokens_y["attention_mask_test"].append(
                list(np.ones_like(tokens_y["decoder_input_ids_test"][-1]))
            )

    tokens_x = padding_sequence(tokens_x, max_len_x)
    tokens_y = padding_sequence(tokens_y, max_len_y)

    return tokens_x, tokens_y


def padding_sequence(tokens, max_len):
    for k in tqdm(tokens.keys()):
        for idx in range(len(tokens[k])):
            if k == "input_ids" or k == "decoder_input_ids":
                tokens[k][idx] = list(
                    np.pad(
                        tokens[k][idx],
                        ((0, max_len - len(tokens[k][idx]))),
                        "constant",
                        constant_values=(1, 1),
                    )
                )
            elif k == "attention_mask":
                tokens[k][idx] = list(
                    np.pad(
                        tokens[k][idx],
                        ((0, max_len - len(tokens[k][idx]))),
                        "constant",
                        constant_values=(0, 0),
                    )
                )
            elif k == "labels":
                tokens[k][idx] = np.pad(
                    tokens[k][idx],
                    ((0, max_len - len(tokens[k][idx]))),
                    "constant",
                    constant_values=(1, 1),
                )
                tokens[k][idx][tokens[k][idx] == 1] = -100
                tokens[k][idx] = list(tokens[k][idx].astype(int))
    return tokens


def prepare_trainer_dataset(
    tokenizer, text_path=None, prefix_mention_is=False, evaluate=False
):

    if not evaluate:
        if not os.path.exists(os.path.join(text_path, "train.source.token.json")):
            encode_data_to_json(os.path.join(text_path, "train"), tokenizer)
        train_tokens_x, train_tokens_y = read_ids_from_json(
            os.path.join(text_path, "train"), prefix_mention_is=prefix_mention_is
        )
        train_set = MedMentionsDataset(train_tokens_x, train_tokens_y)
        return train_set, None, None

    else:
        if not os.path.exists(os.path.join(text_path, "dev.source.token.json")):
            encode_data_to_json(os.path.join(text_path, "dev"), tokenizer)
        if not os.path.exists(os.path.join(text_path, "test.source.token.json")):
            encode_data_to_json(os.path.join(text_path, "test"), tokenizer)

        dev_tokens_x, dev_tokens_y = read_ids_from_json(
            os.path.join(text_path, "dev"), prefix_mention_is=prefix_mention_is
        )
        test_tokens_x, test_tokens_y = read_ids_from_json(
            os.path.join(text_path, "test"), prefix_mention_is=prefix_mention_is
        )

        dev_set = MedMentionsDataset(dev_tokens_x, dev_tokens_y, test_set=True)
        test_set = MedMentionsDataset(test_tokens_x, test_tokens_y, test_set=True)

        return None, dev_set, test_set


class BioGenELDataModule(pl.LightningDataModule):
    def __init__(
        self,
        tokenizer,
        dataset_name: str,
        ontology: BiomedicalOntology,
        save_data_dir="data/",
        preprocess_data=True,
        prefix_mention_is=False,
        evaluate=False,
        resolve_abbrevs=False,
        path_to_abbrev="abbreviations.json",
        batch_size=1,
    ):
        super().__init__()
        self.tokenizer = tokenizer
        self.save_data_dir = save_data_dir
        self.prefix_mention_is = prefix_mention_is
        self.evaluate = evaluate
        self.dataset_name = dataset_name
        self.ontology = ontology
        self.resolve_abbrevs = resolve_abbrevs
        self.batch_size = batch_size
        self.preprocess_data = preprocess_data
        self.path_to_abbrev = path_to_abbrev

    def prepare_data(self):
        # Prepare the .target and .source files
        if self.preprocess_data:
            self.save_data_dir = data_preprocess(
                self.dataset_name,
                self.save_data_dir,
                self.ontology,
                self.path_to_abbrev,
                self.resolve_abbrevs,
            )
            create_trie(self.save_data_dir, use_biobart_tokenizer=True)

    def setup(self, stage=None):
        # Assign train/val datasets for use in dataloaders and prepare the .token.json files
        if stage == "fit" or stage is None:
            if not self.resolve_abbrevs:
                dataset = load_dataset(
                    f"bigbio/{self.dataset_name}", name=f"{self.dataset_name}_bigbio_kb"
                )
                entity_remapping_dict = CUIS_TO_REMAP[self.dataset_name]
                entities_to_exclude = CUIS_TO_EXCLUDE[self.dataset_name]

                if self.dataset_name in VALIDATION_DOCUMENT_IDS:
                    validation_pmids = VALIDATION_DOCUMENT_IDS[self.dataset_name]
                else:
                    validation_pmids = None

                self.deduplicated = dataset_to_df(
                    dataset,
                    entity_remapping_dict=entity_remapping_dict,
                    cuis_to_exclude=entities_to_exclude,
                    val_split_ids=validation_pmids,
                )
                self.deduplicated["mention"] = self.deduplicated["text"]
            else:
                self.deduplicated = load_dataset_df(
                    self.dataset_name, path_to_abbrev=self.path_to_abbrev
                )
                self.deduplicated["mention"] = self.deduplicated["text"]
            if not self.evaluate:
                self.train_set, _, _ = prepare_trainer_dataset(
                    self.tokenizer,
                    self.save_data_dir,
                    self.prefix_mention_is,
                    self.evaluate,
                )
            if self.evaluate:
                _, self.dev_set, self.test_set = prepare_trainer_dataset(
                    self.tokenizer,
                    self.save_data_dir,
                    self.prefix_mention_is,
                    self.evaluate,
                )

    def train_dataloader(self):
        return DataLoader(self.train_set, batch_size=self.batch_size, shuffle=True)

    def dev_dataloader(self):
        return DataLoader(self.dev_set, batch_size=self.batch_size, shuffle=True)

    def test_dataloader(self):
        return DataLoader(self.test_set, batch_size=self.batch_size, shuffle=True)



import os
import io
import json
import pickle
import logger
import numpy as np
import torch

import lightning as L
from transformers import AutoTokenizer
from typing import Optional

from torch.utils.data import DataLoader
import bioel.models.arboel.biencoder.data.data_process as data_process
from bioel.ontology import BiomedicalOntology
from bioel.models.arboel.biencoder.data.data_utils import process_mention_dataset
from bioel.models.arboel.biencoder.model.eval_cluster_linking import (
    filter_by_context_doc_id,
)

import logging

# Configure the logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def read_dataset(dataset_name, preprocessed_json_data_parent_folder, debug=False):
    file_name = "{}.jsonl".format(dataset_name)
    txt_file_path = os.path.join(preprocessed_json_data_parent_folder, file_name)

    samples = []

    with io.open(txt_file_path, mode="r", encoding="utf-8") as file:
        for line in file:
            samples.append(json.loads(line.strip()))
            if debug and len(samples) > 200:
                break

    print(f"Read {len(samples)} samples.")
    return samples


def read_data(split, params, logger):
    """
    Description
    -----------
    Loads dataset samples from a specified path
    Optionally filters out samples without labels
    Checks if the dataset supports multiple labels per sample
    "has_mult_labels" : bool

    Parameters
    ----------
    split : str
        Indicates the portion of the dataset to load ("train", "test", "valid"), used by utils.read_dataset to determine which data to read.
    params : dict(str)
        Contains configuration options
    logger :
        An object used for logging messages about the process, such as the number of samples read.
    """
    samples = read_dataset(split, params["data_path"])
    # Check if dataset has multiple ground-truth labels
    has_mult_labels = "labels" in samples[0].keys()
    if params["filter_unlabeled"]:
        # Filter samples without gold entities
        samples = list(
            filter(
                lambda sample: (
                    (len(sample["labels"]) > 0)
                    if has_mult_labels
                    else (sample["label"] is not None)
                ),
                samples,
            )
        )
    logger.info(f"Read %d {split} samples.." % len(samples))
    return samples, has_mult_labels


"Data module"


class ArboelDataModule(L.LightningDataModule):
    """
    Attributes
    ----------

    - entity_dictionary : list of dict
    Stores the initial and raw entity dictionary
    - train_tensor_data : TensorDataset(context_vecs, label_idxs, n_labels, mention_idx) with :
        - “context_vecs” : tensor containing IDs of (mention + surrounding context) tokens
        - “label_idxs” : tensor with indices pointing to the entities in the entity dictionary that are considered correct labels for the mention.
        - “n_labels” : Number of labels (=entities) associated with the mention
        - “mention_idx” : tensor containing a sequence of integers from 0 to N-1 (N = number of mentions in the dataset) serving as a unique identifier for each mention.
    - train_processed_data : list of dict
    Contains information about mentions (mention_id, mention_name, context, etc…)
    - valid_tensor_data : TensorDataset
    Same as "train_tensor_dataset" but for validation set
    - max_gold_cluster_len : int
    Maximum length of clusters inside gold_cluster
    - train_context_doc_ids : list
    # Store the context_doc_id (=context document indice) for every mention in the train set
    """

    def __init__(self, params):
        """
        Parameters
        ----------
        - params : dict(str)
        Contains configuration options
        """
        super().__init__()
        self.save_hyperparameters(params)

        self.data_path = self.hparams.data_path

        self.tokenizer = AutoTokenizer.from_pretrained(
            self.hparams["model_name_or_path"]
        )

        self.batch_size = self.hparams.get("train_batch_size")

        self.train_processed_data = None
        self.valid_processed_data = None
        self.test_processed_data = None
        self.train_tensor_data = None
        self.valid_tensor_data = None
        self.test_tensor_data = None
        self.train_samples = None
        self.valid_samples = None
        self.test_samples = None
        self.entity_dict_vecs = None

    def prepare_data(self):
        """
        Use this to download and prepare data.
        - Create the entity data files: dictionary.pickle
        - Create the mentions data files:  train.jsonl, valid.jsonl, test.jsonl
        """

        # path to a file where the training data is stored
        self.train_data = os.path.join(self.data_path, "train.jsonl")

        # if the full path to file exist, no need to prepare them, they are already ready
        if not os.path.isfile(self.train_data):

            # Load the ontology object
            with open(
                os.path.join(
                    self.hparams["data_path"],
                    f"{self.hparams['ontology']}_object.pickle",
                ),
                "rb",
            ) as f:
                self.ontology_object = pickle.load(f)

            process_mention_dataset(
                ontology=self.ontology_object,
                dataset=self.hparams["dataset"],
                data_path=self.data_path,
                path_to_abbrev=self.hparams["path_to_abbrev"],
                tax2name_filepath=self.hparams["tax2name_filepath"],
            )

    def setup(self, stage=None):
        """
        For processing and splitting. Called at the beginning of fit (train + validate), test, or predict.
        """

        "entity dictionary"
        # if entity dictionary already tokenized (=add tokens and idx keys), load it
        self.entity_dictionary_pkl_path = os.path.join(
            self.data_path, "entity_dictionary.pickle"
        )
        self.entity_dictionary_loaded = False
        if os.path.isfile(self.entity_dictionary_pkl_path):
            print("Loading stored processed entity dictionary...")
            with open(self.entity_dictionary_pkl_path, "rb") as read_handle:
                self.entity_dictionary = pickle.load(read_handle)
            self.entity_dictionary_loaded = True

        else:  # else load the one not processed yet
            with open(
                os.path.join(self.data_path, "dictionary.pickle"), "rb"
            ) as read_handle:
                self.entity_dictionary = pickle.load(read_handle)

        "Entity dict : drop entity for discovery"
        # For discovery experiment: Drop entities used in training that were dropped randomly from dev/test set
        if self.hparams["drop_entities"]:
            assert self.entity_dictionary
            drop_set_path = (
                self.hparams["drop_set"]
                if self.hparams["drop_set"] is not None
                else os.path.join(self.data_path, "drop_set_mention_data.pickle")
            )
            if not os.path.isfile(drop_set_path):
                raise ValueError(
                    "Invalid or no --drop_set path provided to dev/test mention data"
                )
            with open(drop_set_path, "rb") as read_handle:
                drop_set_data = pickle.load(read_handle)
            # gold cuis indices for each mention in drop_set_data
            drop_set_mention_gold_cui_idxs = list(
                map(lambda x: x["label_idxs"][0], drop_set_data)
            )
            # Make the set unique
            ents_in_data = np.unique(drop_set_mention_gold_cui_idxs)
            # % of drop
            ent_drop_prop = 0.1
            logger.info(
                f"Dropping {ent_drop_prop*100}% of {len(ents_in_data)} entities found in drop set"
            )
            # Number of entity indices to drop
            n_ents_dropped = int(ent_drop_prop * len(ents_in_data))
            # Random selection drop
            rng = np.random.default_rng(seed=17)
            # Indices of all entities that are dropped
            dropped_ent_idxs = rng.choice(
                ents_in_data, size=n_ents_dropped, replace=False
            )

            # Drop entities from dictionary (subsequent processing will automatically drop corresponding mentions)
            keep_mask = np.ones(len(self.entity_dictionary), dtype="bool")
            keep_mask[dropped_ent_idxs] = False
            self.entity_dictionary = np.array(self.entity_dictionary)[keep_mask]

        # fit
        if stage == "fit" or stage is None:
            "training mention data"
            # path to a file where the training data, already processed into tensors is saved
            self.train_tensor_data_pkl_path = os.path.join(
                self.data_path, "train_tensor_data.pickle"
            )
            # path to a file where metadata / additional information about the training data is stored
            self.train_processed_data_pkl_path = os.path.join(
                self.data_path, "train_processed_data.pickle"
            )

            # if the full path to file exist, load the file
            if os.path.isfile(self.train_tensor_data_pkl_path) and os.path.isfile(
                self.train_processed_data_pkl_path
            ):
                print("Loading stored processed train data...")
                with open(self.train_tensor_data_pkl_path, "rb") as read_handle:
                    self.train_tensor_data = pickle.load(read_handle)
                with open(self.train_processed_data_pkl_path, "rb") as read_handle:
                    self.train_processed_data = pickle.load(read_handle)

            else:  # Load and Process train data if not done yet
                # train_samples = list of dict. Each dict contains information about a mention (id, name, context, etc…).
                # Each key can have a dictionary itself. Ex : mention["context"]["tokens"] or mention["context"]["ids"]
                self.train_samples, self.train_mult_labels = read_data(
                    "train", self.hparams, logger
                )

                # train_processed_data = (mention + surrounding context) tokens
                # entity_dictionary = tokenized entities
                # tensor_train_dataset = Dataset containing several tensors (IDs of mention + context / indices of correct entities etc..) # Go check "process_mention_data" for more info
                (
                    self.train_processed_data,
                    self.entity_dictionary,
                    self.train_tensor_data,
                ) = data_process.process_mention_data(
                    samples=self.train_samples,
                    entity_dictionary=self.entity_dictionary,
                    tokenizer=self.tokenizer,
                    max_context_length=self.hparams["max_context_length"],
                    max_cand_length=self.hparams["max_cand_length"],
                    silent=self.hparams["silent"],
                    dictionary_processed=self.entity_dictionary_loaded,
                    context_key=self.hparams["context_key"],
                    multi_label_key="labels" if self.train_mult_labels else None,
                    logger=logger,
                    debug=self.hparams["debug"],
                    knn=self.hparams["knn"],
                )

                # Save the entity dictionary if not already done
                if not self.entity_dictionary_loaded:
                    print("Saving entity dictionary...")
                    with open(self.entity_dictionary_pkl_path, "wb") as write_handle:
                        pickle.dump(
                            self.entity_dictionary,
                            write_handle,
                            protocol=pickle.HIGHEST_PROTOCOL,
                        )

                self.entity_dictionary_loaded = True

                print("Saving processed train data...")
                with open(self.train_tensor_data_pkl_path, "wb") as write_handle:
                    pickle.dump(
                        self.train_tensor_data,
                        write_handle,
                        protocol=pickle.HIGHEST_PROTOCOL,
                    )
                with open(self.train_processed_data_pkl_path, "wb") as write_handle:
                    pickle.dump(
                        self.train_processed_data,
                        write_handle,
                        protocol=pickle.HIGHEST_PROTOCOL,
                    )

            # Prepare tensor containing only ID of (mention + surrounding context) tokens of training set'
            self.train_men_vecs = self.train_tensor_data[:][0]

            "Validation mention data"
            self.valid_tensor_data_pkl_path = os.path.join(
                self.data_path, "valid_tensor_data.pickle"
            )
            self.valid_processed_data_pkl_path = os.path.join(
                self.data_path, "valid_processed_data.pickle"
            )
            # Same as training data :
            # if the full path to file exist, load the file
            if os.path.isfile(self.valid_tensor_data_pkl_path) and os.path.isfile(
                self.valid_processed_data_pkl_path
            ):
                print("Loading stored processed valid data...")
                with open(self.valid_tensor_data_pkl_path, "rb") as read_handle:
                    self.valid_tensor_data = pickle.load(read_handle)
                with open(self.valid_processed_data_pkl_path, "rb") as read_handle:
                    self.valid_processed_data = pickle.load(read_handle)

            else:
                # Load and Process validation data if not done yet
                self.valid_samples, self.valid_mult_labels = read_data(
                    "valid", self.hparams, logger
                )
                self.valid_processed_data, _, self.valid_tensor_data = (
                    data_process.process_mention_data(
                        samples=self.valid_samples,
                        entity_dictionary=self.entity_dictionary,
                        tokenizer=self.tokenizer,
                        max_context_length=self.hparams["max_context_length"],
                        max_cand_length=self.hparams["max_cand_length"],
                        silent=self.hparams["silent"],
                        context_key=self.hparams["context_key"],
                        multi_label_key="labels" if self.valid_mult_labels else None,
                        logger=logger,
                        debug=self.hparams["debug"],
                        knn=self.hparams["knn"],
                        dictionary_processed=self.entity_dictionary_loaded,
                    )
                )

                # Save the entity dictionary if not already done
                if not self.entity_dictionary_loaded:
                    print("Saving entity dictionary...")
                    with open(self.entity_dictionary_pkl_path, "wb") as write_handle:
                        pickle.dump(
                            self.entity_dictionary,
                            write_handle,
                            protocol=pickle.HIGHEST_PROTOCOL,
                        )

                self.entity_dictionary_loaded = True

                print("Saving processed valid data...")
                with open(self.valid_tensor_data_pkl_path, "wb") as write_handle:
                    pickle.dump(
                        self.valid_tensor_data,
                        write_handle,
                        protocol=pickle.HIGHEST_PROTOCOL,
                    )
                with open(self.valid_processed_data_pkl_path, "wb") as write_handle:
                    pickle.dump(
                        self.valid_processed_data,
                        write_handle,
                        protocol=pickle.HIGHEST_PROTOCOL,
                    )
            # Prepare tensor containing only ID of (mention + surrounding context) tokens of validation data
            self.valid_men_vecs = self.valid_tensor_data[:][0]

            # Get clusters of mentions that map to a gold entity
            self.train_gold_clusters = data_process.compute_gold_clusters(
                self.train_processed_data
            )
            # Maximum length of clusters inside gold_cluster
            self.max_gold_cluster_len = 0
            for ent in self.train_gold_clusters:
                if len(self.train_gold_clusters[ent]) > self.max_gold_cluster_len:
                    self.max_gold_cluster_len = len(self.train_gold_clusters[ent])

        # test
        if stage == "test" or stage is None:

            if self.hparams.get("transductive"):
                "training mention data"
                # path to a file where the training data, already processed into tensors is saved
                self.train_tensor_data_pkl_path = os.path.join(
                    self.data_path, "train_tensor_data.pickle"
                )
                # path to a file where metadata / additional information about the training data is stored
                self.train_processed_data_pkl_path = os.path.join(
                    self.data_path, "train_processed_data.pickle"
                )

                # if the full path to file exist, load the file
                if os.path.isfile(self.train_tensor_data_pkl_path) and os.path.isfile(
                    self.train_processed_data_pkl_path
                ):
                    print("Loading stored processed train data...")
                    with open(self.train_tensor_data_pkl_path, "rb") as read_handle:
                        self.train_tensor_data = pickle.load(read_handle)
                    with open(self.train_processed_data_pkl_path, "rb") as read_handle:
                        self.train_processed_data = pickle.load(read_handle)

                else:  # Load and Process train data if not done yet
                    # train_samples = list of dict. Each dict contains information about a mention (id, name, context, etc…).
                    # Each key can have a dictionary itself. Ex : mention["context"]["tokens"] or mention["context"]["ids"]
                    self.train_samples, self.train_mult_labels = read_data(
                        "train", self.hparams, logger
                    )

                    # train_processed_data = (mention + surrounding context) tokens
                    # entity_dictionary = tokenized entities
                    # tensor_train_dataset = Dataset containing several tensors (IDs of mention + context / indices of correct entities etc..) # Go check "process_mention_data" for more info
                    (
                        self.train_processed_data,
                        self.entity_dictionary,
                        self.train_tensor_data,
                    ) = data_process.process_mention_data(
                        samples=self.train_samples,
                        entity_dictionary=self.entity_dictionary,
                        tokenizer=self.tokenizer,
                        max_context_length=self.hparams["max_context_length"],
                        max_cand_length=self.hparams["max_cand_length"],
                        silent=self.hparams["silent"],
                        dictionary_processed=self.entity_dictionary_loaded,
                        context_key=self.hparams["context_key"],
                        multi_label_key="labels" if self.train_mult_labels else None,
                        logger=logger,
                        debug=self.hparams["debug"],
                        knn=self.hparams["knn"],
                    )

                    # Save the entity dictionary if not already done
                    if not self.entity_dictionary_loaded:
                        print("Saving entity dictionary...")
                        with open(
                            self.entity_dictionary_pkl_path, "wb"
                        ) as write_handle:
                            pickle.dump(
                                self.entity_dictionary,
                                write_handle,
                                protocol=pickle.HIGHEST_PROTOCOL,
                            )

                    self.entity_dictionary_loaded = True

                    print("Saving processed train data...")
                    with open(self.train_tensor_data_pkl_path, "wb") as write_handle:
                        pickle.dump(
                            self.train_tensor_data,
                            write_handle,
                            protocol=pickle.HIGHEST_PROTOCOL,
                        )
                    with open(self.train_processed_data_pkl_path, "wb") as write_handle:
                        pickle.dump(
                            self.train_processed_data,
                            write_handle,
                            protocol=pickle.HIGHEST_PROTOCOL,
                        )

                # Prepare tensor containing only ID of (mention + surrounding context) tokens of training set'
                self.train_men_vecs = self.train_tensor_data[:][0]

            "test mention data"
            self.test_tensor_data_pkl_path = os.path.join(
                self.data_path, "test_tensor_data.pickle"
            )
            self.test_processed_data_pkl_path = os.path.join(
                self.data_path, "test_processed_data.pickle"
            )
            # Same as training data :
            # if the full path to file exist, load the file
            if os.path.isfile(self.test_tensor_data_pkl_path) and os.path.isfile(
                self.test_processed_data_pkl_path
            ):
                print("Loading stored processed test data...")
                with open(
                    self.test_tensor_data_pkl_path, "rb"
                ) as read_handle:  # CC7 'rb' = binary read mode
                    self.test_tensor_data = pickle.load(read_handle)
                with open(self.test_processed_data_pkl_path, "rb") as read_handle:
                    self.test_processed_data = pickle.load(read_handle)

            else:
                # Load and Process test data if not done yet
                self.test_samples, self.test_mult_labels = read_data(
                    "test", self.hparams, logger
                )
                self.test_processed_data, _, self.test_tensor_data = (
                    data_process.process_mention_data(
                        samples=self.test_samples,
                        entity_dictionary=self.entity_dictionary,
                        tokenizer=self.tokenizer,
                        max_context_length=self.hparams["max_context_length"],
                        max_cand_length=self.hparams["max_cand_length"],
                        context_key=self.hparams["context_key"],
                        multi_label_key="labels" if self.test_mult_labels else None,
                        silent=self.hparams["silent"],
                        logger=logger,
                        debug=self.hparams["debug"],
                        knn=self.hparams["knn"],
                        dictionary_processed=self.entity_dictionary_loaded,
                    )
                )

                print("Saving processed test data...")
                with open(self.test_tensor_data_pkl_path, "wb") as write_handle:
                    pickle.dump(
                        self.test_tensor_data,
                        write_handle,
                        protocol=pickle.HIGHEST_PROTOCOL,
                    )
                with open(self.test_processed_data_pkl_path, "wb") as write_handle:
                    pickle.dump(
                        self.test_processed_data,
                        write_handle,
                        protocol=pickle.HIGHEST_PROTOCOL,
                    )

            # Prepare tensor containing only ID of (mention + surrounding context) tokens of validation data'
            self.test_men_vecs = self.test_tensor_data[:][0]

        # Store the IDs of the entity in entity_dictionary # It's the equivalent of train_men_vecs for entities
        # (Done here because data_process.process_mention_data will tokenize the entities in entity_dict)
        self.entity_dict_vecs = torch.tensor(
            list(map(lambda x: x["ids"], self.entity_dictionary)), dtype=torch.long
        )

        # Within_doc search
        # Consider if it’s within_doc (=search only within the document)'
        self.train_context_doc_ids = None
        self.valid_context_doc_ids = None
        self.test_context_doc_ids = None

        if self.hparams["within_doc"]:
            logger.info("within_doc")
            # Store the context_doc_id for every mention in the train and valid sets
            if self.train_samples is None:
                self.train_samples, _ = read_data("train", self.hparams, logger)
            self.train_context_doc_ids = [
                s["context_doc_id"] for s in self.train_samples
            ]
            if self.valid_samples is None:
                self.valid_samples, _ = read_data("valid", self.hparams, logger)
            self.valid_context_doc_ids = [
                s["context_doc_id"] for s in self.valid_samples
            ]
            if self.test_samples is None:
                self.test_samples, _ = read_data("test", self.hparams, logger)
            self.test_context_doc_ids = [s["context_doc_id"] for s in self.test_samples]

    def train_dataloader(self):
        # Return the training DataLoader
        return DataLoader(
            self.train_tensor_data,
            batch_size=self.batch_size,
            shuffle=True,
            drop_last=True,
            num_workers=11,
        )

    def val_dataloader(self):
        # Return the validation DataLoader
        return DataLoader(
            self.valid_tensor_data,
            batch_size=self.batch_size,
            num_workers=11,
        )

    def test_dataloader(self):
        # Return the validation DataLoader
        return DataLoader(
            self.test_tensor_data,
            batch_size=self.batch_size,
            num_workers=11,
        )


def main():

    model = "arboel"

    dataset = "ncbi_disease"
    abs_path = "/home2/cye73/data_test2"
    data_path = os.path.join(abs_path, model, dataset)
    os.makedirs(data_path, exist_ok=True)

    # dataset = "bc5cdr"
    # abs_path = "/home2/cye73/data"
    # data_path = os.path.join(abs_path, model, dataset)

    # dataset = "medmentions_st21pv"
    # abs_path = "/home2/cye73/data"
    # data_path = os.path.join(abs_path, model, dataset)

    params_test = {
        "data_path": data_path,
        "batch_size": 64,
        "max_context_length": 128,
        "max_cand_length": 128,
        "context_key": "context",
        "debug": False,
        "knn": 4,
        "model_name_or_path": "dmis-lab/biobert-base-cased-v1.2",
        "out_dim": 768,
        "pull_from_layer": 11,
        "add_linear": True,
        "use_types": True,
        "force_exact_search": True,
        "probe_mult_factor": 1,
        "embed_batch_size": 768,
        "drop_entities": False,
        "within_doc": True,
        "filter_unlabeled": False,
        "model": "arboel",
        # "ontology": "umls",
        # "dataset": "medmentions_st21pv",
        # "ontology_dir": "/mitchell/entity-linking/2017AA/META/",
        "ontology": "medic",
        "dataset": "ncbi_disease",
        "ontology_dir": "/mitchell/entity-linking/kbs/medic.tsv",
        # "ontology": "mesh",
        # "dataset": "bc5cdr",
        # "ontology_dir": "/mitchell/entity-linking/2017AA/META/",
        "silent": False,
        "abbrevs": True,
        "path_to_abbrev": "/home2/cye73/data_test2/abbreviations.json",
    }

    data_module = ArboelDataModule(params=params_test)

    data_module.prepare_data()

    data_module.setup()


if __name__ == "__main__":
    main()



import pickle
import ujson
import os

import csv

import numpy as np

from tqdm import tqdm
from typing import Optional

from bioel.utils.umls_utils import UmlsMappings
from bioel.utils.bigbio_utils import (
    CUIS_TO_REMAP,
    CUIS_TO_EXCLUDE,
    VALIDATION_DOCUMENT_IDS,
    dataset_to_documents,
    dataset_to_df,
)

from bioel.utils.bigbio_utils import (
    load_bigbio_dataset,
    resolve_abbreviation,
    add_deabbreviations,
)
from bioel.ontology import BiomedicalOntology


def get_type_gcd(types, type2geneology, cached_types: dict = None):
    """
    Find the most granular single parent type for an entity with multiple types
    If an entity is assigned multiple types with disjoint type hierarchies
    """
    # Initialize cached_types if it is None
    if cached_types is None:
        cached_types = {}

    if len(types) == 1:
        t = types[0]
        if t not in cached_types:
            cached_types[t] = type2geneology[t][-1]
        return cached_types[t]

    type_tuple = tuple(sorted(types))
    if type_tuple in cached_types:
        return cached_types[type_tuple]
    else:
        geneologies = [type2geneology[t] for t in types]
        min_len = min([len(x) for x in geneologies])
        arr = np.array([gen[:min_len] for gen in geneologies])
        unique_types = np.unique(arr[:, 0])
        if len(unique_types) > 1:
            output_types = ", ".join(list(set([x[-1] for x in geneologies])))
            cached_types[type_tuple] = output_types
            return output_types
        else:
            for i in np.arange(1, arr.shape[1]):
                curr_unique = np.unique(arr[:, i])

                if len(curr_unique) > 1 or i >= 3:
                    cached_types[type_tuple] = unique_types[0]
                    return unique_types[0]
                else:
                    unique_types = curr_unique


def process_ontology(
    ontology: BiomedicalOntology, data_path: str, tax2name_filepath: str = None
):
    """
    This function prepares the entity data : dictionary.pickle

    Parameters
    ----------
    - ontology : str (only umls for now)
        Ontology associated with the dataset
    - data_path : str
        Path where to load and save dictionary.pickle
    - tax2name_filepath : str
        Path to the taxonomy to name file
    """

    # Check if equivalent CUIs are present for the first entity
    first_entity_cui = next(iter(ontology.entities))
    equivalant_cuis = bool(ontology.entities[first_entity_cui].equivalant_cuis)
    print("equivalant cuis :", equivalant_cuis)
    "If dictionary already processed, load it else process and load it"
    entity_dictionary_pkl_path = os.path.join(data_path, "dictionary.pickle")

    if os.path.isfile(entity_dictionary_pkl_path):
        print("Loading stored processed entity dictionary...")
        with open(entity_dictionary_pkl_path, "rb") as read_handle:
            entities = pickle.load(read_handle)

        return entities, equivalant_cuis

    if tax2name_filepath:
        with open(tax2name_filepath, "r") as f:
            tax2name = ujson.load(f)

    ontology_entities = []
    for cui, entity in tqdm(ontology.entities.items()):
        new_entity = {}

        if ontology.name.lower() in ["umls"]:
            with open(os.path.join(data_path, "tui2type_hierarchy.json"), "r") as f:
                type2geneology = ujson.load(f)
            entity.types = get_type_gcd(entity.types, type2geneology)

        new_entity["cui"] = entity.cui
        new_entity["title"] = entity.name
        new_entity["types"] = f"{entity.types}"

        if entity.aliases:
            if entity.definition:
                if entity.taxonomy:
                    new_entity["description"] = (
                        f"{entity.name} ( {tax2name[str(entity.taxonomy)]}, {entity.types} : {entity.aliases} ) [{entity.definition}]"
                    )

                else:
                    new_entity["description"] = (
                        f"{entity.name} ( {entity.types} : {entity.aliases} ) [{entity.definition}]"
                    )

            else:
                if entity.taxonomy:
                    new_entity["description"] = (
                        f"{entity.name} ( {tax2name[str(entity.taxonomy)]}, {entity.types} : {entity.aliases} )"
                    )
                else:
                    new_entity["description"] = (
                        f"{entity.name} ( {entity.types} : {entity.aliases} )"
                    )

        else:
            if entity.definition:
                if entity.taxonomy:
                    new_entity["description"] = (
                        f"{entity.name} ( {tax2name[str(entity.taxonomy)]}, {entity.types}) [{entity.definition}]"
                    )

                else:
                    new_entity["description"] = (
                        f"{entity.name} ( {entity.types}) [{entity.definition}]"
                    )
            else:
                if entity.taxonomy:
                    new_entity["description"] = (
                        f"{entity.name} ( {tax2name[str(entity.taxonomy)]}, {entity.types})"
                    )
                else:
                    new_entity["description"] = f"{entity.name} ({entity.types})"

        if hasattr(entity, "metadata") and entity.metadata:
            new_entity["description"] += f" {entity.metadata}"

        if equivalant_cuis:
            new_entity["cuis"] = entity.equivalant_cuis

        ontology_entities.append(new_entity)

    # Check if the directory exists, and create it if it does not
    if not os.path.exists(data_path):
        os.makedirs(data_path)

    # Save entities to pickle file
    with open(os.path.join(data_path, "dictionary.pickle"), "wb") as f:
        pickle.dump(ontology_entities, f)

    entities = pickle.load(open(os.path.join(data_path, "dictionary.pickle"), "rb"))
    return entities, equivalant_cuis


# Function for preparing the mentions in the dataset into the right format for each model
def process_mention_dataset(
    ontology: BiomedicalOntology,
    dataset: str,
    data_path: str,
    path_to_abbrev=None,
    tax2name_filepath=None,
):
    """
    This function prepares the mentions data :  Creates the train.jsonl, valid.jsonl, test.jsonl
    Each .jsonl contains data in the following format :
    {'mention': mention,
    'mention_id': ID of the mention, (optional)
    'context_left': context before mention,
    'context_right': context after mention,
    'context_doc_id': ID of the doc, (optional)
    'type': type
    'label_id': label_id,
    'label': entity description, (optional)
    'label_title': entity title

    Parameters
    ----------
    - ontology : BiomedicalOntology object
    Ontology associated with the dataset
    - dataset : str
    Name of the dataset
    - data_path : str
    Path where to load and save dictionary.pickle
    - tax2name_filepath : str
    Path to the taxonomy to name file
    """
    data = load_bigbio_dataset(dataset_name=dataset)
    exclude = CUIS_TO_EXCLUDE[dataset]
    remap = CUIS_TO_REMAP[dataset]

    if path_to_abbrev:
        data = add_deabbreviations(dataset=data, path_to_abbrev=path_to_abbrev)

    entities, equivalant_cuis = process_ontology(ontology, data_path, tax2name_filepath)

    entity_dictionary = {d["cui"]: d for d in tqdm(entities)}  # CC1

    # For ontology with multiples cuis
    if equivalant_cuis:
        # Need to redo this since we have multiple synonymous CUIs for ncbi_disease
        entity_dictionary = {cui: d for d in tqdm(entities) for cui in d["cuis"]}
        cui_synsets = {}
        for subdict in tqdm(entities):
            for cui in subdict["cuis"]:
                if cui in subdict:
                    print(cui, cui_synsets[cui], subdict["cuis"])
                cui_synsets[cui] = subdict["cuis"]
        with open(os.path.join(data_path, "cui_synsets.json"), "w") as f:
            f.write(ujson.dumps(cui_synsets, indent=2))

    if dataset in VALIDATION_DOCUMENT_IDS:
        validation_pmids = VALIDATION_DOCUMENT_IDS[dataset]
    else:
        print("ERROR!!!")

    # Convert BigBio dataset to pandas DataFrame
    df = dataset_to_df(
        data,
        entity_remapping_dict=remap,
        cuis_to_exclude=exclude,
        val_split_ids=validation_pmids,
    )
    # Return dictionary of documents in BigBio dataset
    docs = dataset_to_documents(data)
    label_len = df["db_ids"].map(lambda x: len(x)).max()
    print("Max labels on one doc:", label_len)

    abbrev_dict = ujson.load(open(path_to_abbrev))

    for split in df.split.unique():
        ents_in_split = []
        for d in tqdm(
            df.query("split == @split").to_dict(orient="records"),
            desc=f"Creating correct mention format for {split} dataset",
        ):
            abbrev_resolved = False
            offsets = d["offsets"]
            doc_id = d["document_id"]
            doc = docs[doc_id]
            mention = d["text"]

            # Resolve abbreviaions if desired
            if abbrev_dict is not None:
                deabbreviated_mention = resolve_abbreviation(
                    doc_id, mention, abbrev_dict
                )
                abbrev_resolved = True

            # Get offsets and context
            start = offsets[0][0]  # start on the mention
            end = offsets[-1][-1]  # end of the mention
            before_context = doc[:start]  # left context
            after_context = doc[end:]  # right context

            # ArboEL can't handle multi-labels, so we randomly choose one.
            if len(d["db_ids"]) == 1:
                label_id = d["db_ids"][0]

            # For ontology with multiples cuis
            elif equivalant_cuis:
                labels = []
                used_cuis = set([])
                choosable_ids = []
                for db_id in d["db_ids"]:
                    if db_id in used_cuis:
                        continue
                    else:
                        used_cuis.update(set(entity_dictionary[db_id]["cuis"]))
                    choosable_ids.append(db_id)

                label_id = np.random.choice(choosable_ids)

            else:
                label_id = np.random.choice(d["db_ids"])

            # Check if we missed something
            if label_id not in entity_dictionary:
                # print(label_id)
                continue

            output = [
                {
                    "mention": mention,
                    "mention_id": d["mention_id"],
                    "context_left": before_context,
                    "context_right": after_context,
                    "context_doc_id": doc_id,
                    "type": d["type"][0],
                    "label_id": label_id,
                    "label": entity_dictionary[label_id]["description"],
                    "label_title": entity_dictionary[label_id]["title"],
                }
            ]

            if abbrev_resolved:
                output.append(
                    {
                        "mention": deabbreviated_mention,
                        "mention_id": d["mention_id"] + ".abbr_resolved",
                        "context_left": before_context,
                        "context_right": after_context,
                        "context_doc_id": doc_id,
                        "type": d["type"][0],
                        "label_id": label_id,
                        "label": entity_dictionary[label_id]["description"],
                        "label_title": entity_dictionary[label_id]["title"],
                    }
                )

            ents_in_split.extend(output)

        split_name = split
        if split == "validation":
            split_name = "valid"
        with open(os.path.join(data_path, f"{split_name}.jsonl"), "w") as f:
            f.write("\n".join([ujson.dumps(x) for x in ents_in_split]))



from typing import List, Optional, Union, Dict
import time
import logging
import json
from dataclasses import dataclass
import ujson

import torch
from torch import Tensor as T
from transformers import PreTrainedTokenizer
from bioel.utils.bigbio_utils import (
    load_bigbio_dataset,
    add_deabbreviations,
    dataset_to_df,
    resolve_abbreviation,
    dataset_to_documents,
    get_left_context,
    get_right_context,
)
from bioel.utils.dataset_consts import (
    DATASET_NAMES,
    CUIS_TO_EXCLUDE,
    CUIS_TO_REMAP,
)
from tqdm.auto import tqdm

tqdm.pandas()


logger = logging.getLogger()


@dataclass
class Mention:
    cui: str
    start: int
    end: int
    text: str
    types: List[str]
    deabbreviated_text: Optional[str]


class Dataset(torch.utils.data.Dataset):
    def __init__(
        self,
        dataset_name: str,
        splits: Optional[List[str]] = ["train", "valid", "test"],
        abbreviations_path=None,
    ):
        """
        Parameters
        ---------
        - dataset_name: str
        Name of the dataset to load. Example : "bc5cdr", "ncbi-disease", etc...
        - splits: List[str]
        List of splits to load. Default: ["train", "valid", "test"]
        - abbreviations_path: str
        Path to the abbreviations dictionary. Default: None
        """
        self.dataset_name = dataset_name
        self.data = load_bigbio_dataset(dataset_name)
        self.splits = splits
        self.abbreviations_path = abbreviations_path
        self.name_to_cuis = {}

        exclude = CUIS_TO_EXCLUDE[dataset_name]
        remap = CUIS_TO_REMAP[dataset_name]

        if self.abbreviations_path:
            self.data = add_deabbreviations(
                dataset=self.data, path_to_abbrev=self.abbreviations_path
            )
            print("Resolved abbreviations")
        else:
            print("No abbreviations dictionary found.")

        df = dataset_to_df(
            self.data, cuis_to_exclude=exclude, entity_remapping_dict=remap
        )
        df["start"] = df["offsets"].map(lambda x: x[0][0])
        df["end"] = df["offsets"].map(lambda x: x[-1][-1])
        # df = df[df.split.isin(splits)]
        # print(dataset_name, splits, df.split.unique())

        self.df = df

        self.documents = dataset_to_documents(self.data)


if __name__ == "__main__":
    # Testss
    abbreviations_path = "/home2/cye73/data_test/abbreviations.json"
    bc5cdr = Dataset("bc5cdr", ["train", "validation", "test"], abbreviations_path)
    bc5cdr.data



import numpy as np
import seaborn as sns
import ujson
import sys
import time
import itertools
import pickle
import math
import os

from tqdm.auto import tqdm
from collections import defaultdict
from matplotlib import pyplot as plt
import pandas as pd
from scipy.stats import chi2_contingency

from bioel.utils.bigbio_utils import (
    load_bigbio_dataset,
    add_deabbreviations,
    dataset_to_documents,
    dataset_to_df,
    load_dataset_df,
    resolve_abbreviation,
    CUIS_TO_REMAP,
    CUIS_TO_EXCLUDE,
    DATASET_NAMES,
    VALIDATION_DOCUMENT_IDS,
)
from bioel.utils.dataset_consts import (
    dataset_to_pretty_name,
    model_to_pretty_name,
    model_to_color,
)


def metamap_to_candidates(metamap_output, mappings_first=False):
    """
    Converts MetaMap output into a dictionary mapping text mentions to candidate CUIs.

    Parameters
    ----------
    metamap_output : pandas.DataFrame
        A DataFrame containing MetaMap output with columns for text mentions and their corresponding mapping and candidate CUIs.
    mappings_first : bool, optional
        A flag indicating the order of columns in the DataFrame.
        If True, the columns are ["text", "mapping_cui_list", "candidate_cui_list"].
        If False, the columns are ["text", "candidate_cui_list", "mapping_cui_list"] (default is False).

    Returns
    -------
    dict
        A dictionary where keys are text mentions and values are lists of candidate CUIs.
    """
    if mappings_first:
        cols = ["text", "mapping_cui_list", "candidate_cui_list"]
    else:
        cols = [
            "text",
            "candidate_cui_list",
            "mapping_cui_list",
        ]

    # Create mapping from text to candidates for MedMentions
    text2candidates = defaultdict(list)

    for row in metamap_output[cols].values:
        text = row[0]
        candidates = eval(row[1])
        for c in eval(row[2]):
            candidates
        # TODO: Need to account for correct database for
        candidates = [
            ["UMLS:" + x if ":" not in x else x.replace("ncbigene", "NCBIGene")]
            for x in candidates
        ]
        text2candidates[text] = candidates

    return text2candidates


def deduplicate_candidates(candidates, return_counts=False):
    """
    Deduplicates a list of candidate CUIs and optionally returns their counts.

    Parameters
    ----------
    candidates : list
        A list of candidate CUIs. Each element can be either a string (CUI) or a list of strings (CUIs).
    return_counts : bool, optional
        A flag indicating whether to return the counts of each CUI. Default is False.

    Returns
    -------
    list
        A list of deduplicated CUIs.
    dict (optional)
        A dictionary with CUIs as keys and their counts as values. Only returned if `return_counts` is True.
    """

    deduplicated = []
    counts = defaultdict(int)
    for c in candidates:
        if type(c) == list:
            deduped_subset = [x for x in c if x not in counts]
            if len(deduped_subset) > 0:
                deduplicated.append(deduped_subset)

            for x in c:
                counts[x] += 1

        else:
            counts[c] += 1
            if c not in deduplicated:
                deduplicated.append(c)

    if return_counts:
        return deduplicated, counts

    return deduplicated


def output_list_to_candidates(output_list, k=20, sort_by_counts=False):
    """
    Turn model output into candidate list
    """
    output2candidates = defaultdict(list)
    for mention in output_list:
        doc_id = mention["document_id"]
        offsets = mention["offsets"]
        joined_offsets = ";".join([",".join([str(y) for y in x]) for x in offsets])
        # candidates = [y for x in mention['candidates'] for y in x]
        candidates = mention["candidates"]
        deduplicated_candidates, counts = deduplicate_candidates(
            candidates, return_counts=True
        )

        # Syntax to prioritize candidates that come up more frequently in retrieved results
        if sort_by_counts:
            top_cuis = sorted(
                [(k, v) for k, v in counts.items()], key=lambda x: x[1], reverse=True
            )
            grouped_candidates = []
            for k, g in itertools.groupby(top_cuis, key=lambda x: x[1]):
                grouped_candidates.append(list(g))

            output2candidates[(doc_id, joined_offsets)] = grouped_candidates
        else:
            output2candidates[(doc_id, joined_offsets)] = deduplicated_candidates[:k]

    return output2candidates


def add_candidates_to_df(df, candidate_dict, new_col_name, eval_col="text"):
    """
    Add candidates produced by a model to results dataframe for evaluation

    Parameters:
    ----------------------------
        df: pandas.DataFrame
            Dataframe of bigbio mention produced by bigbio_utils.dataset_to_df

        candidate_dict: dict
            Mapping of (document_id, joined_offsets) -> candidates

        new_col_name: str
            Name of new column

        eval_col: Column to evaluate on
    """
    df[new_col_name] = df[["document_id", "joined_offsets"]].apply(
        lambda x: candidate_dict[(x[0], x[1])], axis=1
    )


def correct_medlinker_candidates(medlinker_output):
    for x in medlinker_output:
        new_candidates = []
        for y in x["candidates"]:
            single_candidate = []
            for c in y:
                if c is None:
                    continue
                if not c.startswith("MESH") and not c.startswith("OMIM"):
                    c = c.replace("ESH", "MESH")
                single_candidate.append(c)
            new_candidates.append(single_candidate)
        x["candidates"] = new_candidates
    return medlinker_output


def list_flatten(nested_list):
    used = set([])
    flattened = []
    for x in nested_list:
        for y in x:
            if y not in used:
                flattened.append(y)
                used.add(y)

    return flattened


def min_hit_index(gold_cuis, candidates, eval_mode):
    """
    Find index of first hit in candidates
    """
    # Check if no candidates
    if candidates == [[]]:
        return 1000000

    if eval_mode == "basic":
        flat_candidates = list_flatten(
            candidates
        )  # Flatten list of lists to find the correct hit index (and not consider several cuis for one test)
        for i, c in enumerate(flat_candidates):
            if c in gold_cuis:
                return i
    elif eval_mode == "strict":
        for i, c in enumerate(candidates):
            if all(x in gold_cuis for x in c):
                return i

    elif eval_mode == "relaxed":
        for i, c in enumerate(candidates):
            if any(x in gold_cuis for x in c):
                return i

    else:
        raise ValueError(f"eval_mode {eval_mode} not supported")

    return 1000000


def recall_at_k(
    df,
    candidate_col,
    gold_col="db_ids",
    max_k: int = 10,
    filter_null=False,
    eval_mode="basic",
):
    """
    Compute recall@k for all values of k < max_k
    """
    # Filter rows wilt null values after CUI remapping
    if filter_null:
        before_row_count = df.shape[0]
        df = df[df[gold_col].map(lambda x: len(x) > 0)]
        after_row_count = df.shape[0]
        print(f"Dropped {before_row_count - after_row_count} rows with null db_ids")

    df[f"{candidate_col}_min_hit_index"] = df[[gold_col, candidate_col]].apply(
        lambda x: min_hit_index(x[0], x[1], eval_mode=eval_mode), axis=1
    )

    recall_at_k_dict = {}
    for k in range(1, max_k + 1):
        recall_at_k_dict[k] = (df[f"{candidate_col}_min_hit_index"] < k).mean()

    return recall_at_k_dict


def plot_recall_at_k(
    recall_dict, max_k=10, legend_key=None, ax=None, color=None, alpha=1
):
    if ax is not None:
        ax.plot(
            recall_dict.keys(),
            recall_dict.values(),
            marker="o",
            label=legend_key,
            color=color,
            alpha=alpha,
        )
    else:
        plt.plot(
            recall_dict.keys(),
            recall_dict.values(),
            marker="o",
            label=legend_key,
            color=color,
            alpha=alpha,
        )
    # if legend_key is not None:
    #     plt.legend()


def precision_at_k_DK(df, hit_column, k):
    """
    Calculates Precision@k for a given DataFrame and value of k. (David's version)
    ---------
    df : pandas.DataFrame
        DataFrame containing the results of the model.
    hit_column : str
        Name of the column containing the hit index.
    k : int
        Value of k.
    """
    hits_within_k = df[hit_column] <= (k - 1)
    precision = hits_within_k.sum() / len(df)
    return precision


def precision_at_k(df, hit_column, k):
    """
    Calculates Precision@k for a given DataFrame and value of k.
    ---------
    df : pandas.DataFrame
        DataFrame containing the results of the model.
    hit_column : str
        Name of the column containing the hit index.
    k : int
        Value of k.
    """
    precision_sum = 0
    for hit_index in df[hit_column]:
        if hit_index < k:
            precision_sum += 1 / k
    precision_at_k = precision_sum / len(df)
    return precision_at_k


def average_precision_at_k(hit_index, k):
    """
    Calculates Average Precision@k for a single query (mention).
    --------
    hit_index : int
        Index of the first hit.
    k : int
    """
    if hit_index <= (k - 1):  # Relevant item is within top k
        return 1 / (
            hit_index + 1
        )  # Precision at the position where the relevant item is found
    else:
        return 0


def mean_average_precision_at_k(df, hit_column, k):
    """
    Calculates MAP@k for the entire DataFrame.
    --------
    df : pandas.DataFrame
        DataFrame containing the results of the model.
    hit_column : str
        Name of the column containing the hit index.
    k : int
    """
    average_precisions = df[hit_column].apply(lambda x: average_precision_at_k(x, k))
    map_k = average_precisions.mean()  # Average over all queries
    return map_k


class Evaluate:
    def __init__(
        self,
        dataset_names,
        model_names,
        path_to_result,
        eval_strategies=["basic", "relaxed", "strict"],
        abbreviations_path=None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        dataset_names : list of str
            A list of dataset names to be used in the evaluation.
        model_names : list of str
            A list of model names to be evaluated.
        path_to_result : nested dict
            The path to the directory where the result files for each model and dataset are stored.
            Ex : path_to_result = {
                "bc5cdr": {
                    "arboEL": "/path/to/output_eval.json",
                    "krissbert": "/path/to/output_eval.json",
                },
                "ncbi_disease": {
                    "arboEL": "/path/to/output_eval.json",
                    "krissbert": "/path/to/output_eval.json",
                },
                "nlmchem" : {
                    "arboEL": "/path/to/output_eval.json",
                    "krissbert": "/path/to/output_eval.json",
                },
            }
        abbreviations_path : str
            The path to the JSON file containing abbreviations to be used for processing the datasets.
        eval_strategies : list of str, optional
            A list of evaluation strategies to be used (default is ["basic", "relaxed", "strict"]).
        """
        self.dataset_names = dataset_names
        self.model_names = model_names
        self.path_to_result = path_to_result
        self.abbreviations_path = abbreviations_path
        self.eval_strategies = eval_strategies
        self.data = {}  # df for each dataset
        self.full_results = defaultdict(dict)
        self.datasets = {}
        self.error_analysis_dfs = {}  # Results with hit_index for error analysis
        self.recall_all_eval_strategies = (
            {}
        )  # Recall@k for all eval strategies, datasets, and models
        self.detailed_results_analysis = (
            {}
        )  # Detailed results for failure stage, statistical significance (p_values, chi2), accuracy per type, recall@k per type, MAP@k
        self.max_k = kwargs.get("max_k", 10)
        self.split = kwargs.get("split", "test")

    def load_results(self):
        """
        Loads the evaluation results for each model and dataset.

        Iterates over each dataset and model, loading the corresponding evaluation results from JSON files.
        Raises a FileNotFoundError if the expected result file for any model and dataset combination does not exist.
        """
        for name in self.dataset_names:
            self.datasets[name] = {}
            for model in self.model_names:
                file_path = self.path_to_result[name][model]
                if file_path and os.path.exists(file_path):
                    self.datasets[name][model] = ujson.load(open(file_path))
                else:
                    print(
                        f"Skipping model {model} for dataset {name} because the file is missing or the path is invalid."
                    )

    def process_datasets(self):
        """
        Processes each dataset by loading and preparing it for evaluation.

        For each dataset, this method:
        - Loads the dataset using `load_bigbio_dataset`.
        - Optionallly applies de-abbreviations to the dataset.
        - Converts the dataset to a DataFrame, applying entity remappings and exclusions.
        - Maps model results to the DataFrame and prepares it for evaluation.
        """
        for name in self.dataset_names:
            data = load_bigbio_dataset(name)
            exclude = CUIS_TO_EXCLUDE[name]
            remap = CUIS_TO_REMAP[name]
            if self.abbreviations_path:
                data_with_abbrev = add_deabbreviations(
                    dataset=data, path_to_abbrev=self.abbreviations_path
                )
                df = dataset_to_df(
                    data_with_abbrev,
                    entity_remapping_dict=remap,
                    cuis_to_exclude=exclude,
                )

            else:
                df = dataset_to_df(
                    data, entity_remapping_dict=remap, cuis_to_exclude=exclude
                )

            df["joined_offsets"] = df.offsets.map(
                lambda offsets: ";".join(
                    [",".join([str(y) for y in x]) for x in offsets]
                )
            )

            for model, results in self.datasets[name].items():
                print(f"dataset : {name}, model : {model}")
                for x in results:
                    x["candidates"] = [y for y in x["candidates"]]

                dict = {
                    x["mention_id"]: {
                        "db_ids": x["db_ids"],
                        "candidates": x["candidates"],
                    }
                    for x in results
                    if "abbr" not in x["mention_id"]
                }
                df[f"{model}"] = df["mention_id"].map(
                    lambda x: dict[x]["candidates"] if x in dict else [[] * 20]
                )

                resolve_abbr_dict = {
                    x["mention_id"].rstrip(".abbr_resolved"): {
                        "db_ids": x["db_ids"],
                        "candidates": x["candidates"],
                    }
                    for x in results
                    if "abbr" in x["mention_id"]
                }
                df[f"{model}_resolve_abbrev"] = df["mention_id"].map(
                    lambda x: (
                        resolve_abbr_dict[x]["candidates"]
                        if x in resolve_abbr_dict
                        else [[] * 20]
                    )
                )

            df = df[df["split"] == self.split].reset_index(drop=True)
            self.data[name] = df

    def evaluate(self, eval_strategies=None):
        """
        Evaluates the performance of models on each dataset using different evaluation strategies.

        For each evaluation strategy, this method:
        - Computes recall metrics for each model and dataset.
        - Handles cases where resolved abbreviations are used.
        - Stores detailed results for error analysis and recall metrics for different evaluation strategies.

        Parameters
        ----------
        eval_strategies : list of str, optional
            A list of evaluation strategies to be used (default is ["basic", "relaxed", "strict"]).

        """
        eval_strategies = eval_strategies if eval_strategies else self.eval_strategies
        self.full_results = defaultdict(dict)  # Initialize as defaultdict
        use_resolved_abbrevs = True if self.abbreviations_path else False
        for eval_strategy in tqdm(eval_strategies):
            all_recall = {}
            print(f"Eval Strategy: {eval_strategy}")

            dfs_one_eval_strategy = {}
            for name in self.dataset_names:
                print(name)
                self.full_results[eval_strategy][name] = self.data[name].copy()
                df = self.full_results[eval_strategy][name]
                recall_dict = {}
                for model in self.model_names:
                    if model in df.columns:
                        if model == "scispacy":
                            recall_dict[model] = recall_at_k(
                                df,
                                model,
                                eval_mode=eval_strategy,
                                gold_col="db_ids",
                                max_k=self.max_k,
                            )
                        else:
                            recall_dict[model] = recall_at_k(
                                df,
                                model,
                                eval_mode=eval_strategy,
                                max_k=self.max_k,
                            )

                        if model + "_resolve_abbrev" in df.columns:
                            cand_col = model + "_resolve_abbrev"
                            if model == "scispacy":
                                deabbrev_recall = recall_at_k(
                                    df,
                                    cand_col,
                                    eval_mode=eval_strategy,
                                    gold_col="db_ids",
                                    max_k=self.max_k,
                                )
                            else:
                                deabbrev_recall = recall_at_k(
                                    df,
                                    cand_col,
                                    eval_mode=eval_strategy,
                                    max_k=self.max_k,
                                )
                            if use_resolved_abbrevs:
                                recall_dict[model] = deabbrev_recall

                all_recall[name] = recall_dict

                hit_index_cols = sorted([x for x in df.columns if "min_hit_index" in x])
                hit_index_df = df[
                    [
                        "document_id",
                        "text",
                        "deabbreviated_text",
                        "type",
                        "db_ids",
                        "mention_id",
                        "joined_offsets",
                    ]
                    + hit_index_cols
                ]
                hit_index_df = hit_index_df.rename(
                    {x: x.replace("_min_hit_index", "") for x in hit_index_df.columns},
                    axis=1,
                )

                dfs_one_eval_strategy[name] = hit_index_df

            self.error_analysis_dfs[eval_strategy] = dfs_one_eval_strategy
            self.recall_all_eval_strategies[eval_strategy] = all_recall

    def plot_results(self, eval_strategies=None):
        """
        Plots recall@k for each model and dataset using different evaluation strategies.

        This method:
        - Dynamically creates a grid of subplots based on the number of datasets.
        - Plots recall@k curves for each model on each dataset.
        - Uses distinct colors for each model for better visualization.
        - Handles different grid sizes and ensures proper labeling and legends.

        Parameters
        ----------
        eval_strategies : list of str, optional
            A list of evaluation strategies to be used (default is ['basic', 'relaxed', 'strict']).

        """
        eval_strategies = eval_strategies if eval_strategies else self.eval_strategies
        # Define a color palette
        palette = sns.color_palette(
            "husl", len(self.model_names)
        )  # Using husl palette for distinct colors
        model_to_color = {model: palette[i] for i, model in enumerate(self.model_names)}

        for eval_strategy in tqdm(eval_strategies):
            print(f"{eval_strategy=}")
            all_recall = self.recall_all_eval_strategies[eval_strategy]

            # Determine the grid size based on the number of datasets
            num_datasets = len(self.dataset_names)
            num_cols = min(
                3, num_datasets
            )  # Limit the number of columns to 3 for better visualization
            num_rows = math.ceil(num_datasets / num_cols)

            fig, axs = plt.subplots(
                num_rows,
                num_cols,
                sharex=True,
                sharey=True,
                figsize=(num_cols * 4, num_rows * 4),
            )

            # Flatten the axs array if it's not 1D
            if num_rows == 1 and num_cols == 1:
                axs = [axs]
            elif num_rows == 1 or num_cols == 1:
                axs = axs.flatten()
            else:
                axs = axs.ravel()

            for ax in axs:
                if num_cols > 1:
                    ax.set_xlabel("k")
                if num_rows > 1:
                    ax.set_ylabel("recall@k")

            for idx, name in enumerate(self.dataset_names):
                recall_dict = all_recall[name]
                df = self.full_results[eval_strategy][name]

                ax = axs[idx]
                ax.title.set_text(dataset_to_pretty_name.get(name, name))

                for model in self.model_names:
                    if model in recall_dict:
                        plot_recall_at_k(
                            recall_dict[model],
                            # legend_key=model_to_pretty_name.get(model, model),
                            legend_key=model,
                            ax=ax,
                            color=model_to_color[model],
                            alpha=0.7,
                            max_k=self.max_k,
                        )

                print(f"{dataset_to_pretty_name.get(name, name)=}")

            # Hide any unused subplots
            for ax in axs[num_datasets:]:
                ax.axis("off")

            handles, labels = axs[0].get_legend_handles_labels()
            fig.legend(
                handles,
                labels,
                loc="lower center",
                ncol=len(self.model_names),
                bbox_to_anchor=(0.5, -0.05),
            )

            fig.suptitle(
                f"Recall@K for all models using {eval_strategy} evaluation strategy",
                fontsize=16,
            )
            plt.show()

    def detailed_results(self, k=None, eval_strategies=None):
        """
        Returns detailed results for further analysis : accuracy per type, recall@k per type, failure stage, statistical significance (p_values, chi2)
        The results are stored in the attribute detailed_results_analysis and is a nested dictionary with the following structure:
        self.detailed_results_analysis[eval_strat][dataset_name][model_name][feature]
        - eval_strat : evaluation strategy (e.g. basic, relaxed, strict)
        - dataset_name : name of the dataset (e.g. bc5cdr, ncbi_disease, nlmchem)
        - model_name : name of the model (e.g. arboEL, krissbert)
        - feature : feature of interest (count_miss_CG_per_type, count_miss_NED_per_type, count_success_per_type, count_hit_k_per_type, mentions_count_per_type, accuracy_per_type, recall_k_per_type, failure_stage_CG, failure_stage_NED, contingency_table_CG, contingency_table_NED, ChiSquare_test_NED, ChiSquare_test_CG)
        """
        k = k if k else self.max_k
        eval_strategies = eval_strategies if eval_strategies else self.eval_strategies
        resolve_abbrev = True if self.abbreviations_path else False
        for eval_strat in eval_strategies:
            self.detailed_results_analysis[eval_strat] = {}
            for dataset_name in self.dataset_names:
                self.detailed_results_analysis[eval_strat][dataset_name] = {}

                for model_name in self.model_names:
                    if (
                        model_name
                        not in self.error_analysis_dfs[eval_strat][
                            dataset_name
                        ].columns  # Check if the model has been evaluated on the dataset
                    ):
                        print(
                            f"Skipping model {model_name} for dataset {dataset_name} because the column doesn't exist."
                        )
                        continue

                    self.detailed_results_analysis[eval_strat][dataset_name][
                        model_name
                    ] = {}

                    df = self.error_analysis_dfs[eval_strat][dataset_name]
                    assert (
                        "type" in df.columns
                    ), "'type' column is missing in the DataFrame for a comparison per type"
                    # Convert the 'type' column to strings
                    df["type"] = df["type"].apply(str)

                    name = model_name
                    if resolve_abbrev:
                        name += "_resolve_abbrev"
                    print(f"Processing {name}")

                    # Number of mentions where failure happened at CG step per type
                    count_miss_CG_per_type = (
                        df[df[name] == 1000000].groupby("type").size()
                    )
                    self.detailed_results_analysis[eval_strat][dataset_name][
                        model_name
                    ][
                        "count_miss_CG_per_type"
                    ] = (
                        count_miss_CG_per_type.to_frame().reset_index()
                    )  # pandas series to dataframe + reset index (not aligned otherwise)

                    # Number of mentions where failure happened at NED step per type
                    count_miss_NED_per_type = (
                        df[(df[name] != 1000000) & (df[name] != 0)]
                        .groupby("type")
                        .size()
                    )
                    self.detailed_results_analysis[eval_strat][dataset_name][
                        model_name
                    ][
                        "count_miss_NED_per_type"
                    ] = (
                        count_miss_NED_per_type.to_frame().reset_index()
                    )  # pandas series to dataframe + reset index (not aligned otherwise)

                    # Number of correctly linked mentions for each type
                    count_success_per_type = df[df[name] == 0].groupby("type").size()
                    self.detailed_results_analysis[eval_strat][dataset_name][
                        model_name
                    ][
                        "count_success_per_type"
                    ] = count_success_per_type.to_frame().reset_index()

                    # Number of mentions with hit_index < k (recall@k) per type
                    count_hit_k_per_type = df[df[name] < k].groupby("type").size()
                    self.detailed_results_analysis[eval_strat][dataset_name][
                        model_name
                    ][
                        "count_hit_k_per_type"
                    ] = count_hit_k_per_type.to_frame().reset_index()

                    # Total number of mentions for each type
                    mentions_count_per_type = df["type"].value_counts()
                    self.detailed_results_analysis[eval_strat][dataset_name][
                        model_name
                    ][
                        "mentions_count_per_type"
                    ] = mentions_count_per_type.to_frame().reset_index()

                    # Accuracy for each type
                    accuracy_per_type = (
                        count_success_per_type / mentions_count_per_type
                    ).fillna(0)
                    self.detailed_results_analysis[eval_strat][dataset_name][
                        model_name
                    ]["accuracy_per_type"] = accuracy_per_type.to_frame().reset_index()

                    # recall@k for each type
                    recall_k_per_type = (
                        count_hit_k_per_type / mentions_count_per_type
                    ).fillna(0)
                    self.detailed_results_analysis[eval_strat][dataset_name][
                        model_name
                    ]["recall_k_per_type"] = recall_k_per_type.to_frame().reset_index()

                    # Failure from CG stage
                    failure_stage_CG = count_miss_CG_per_type.sum() / (
                        count_miss_CG_per_type.sum() + count_miss_NED_per_type.sum()
                    )
                    self.detailed_results_analysis[eval_strat][dataset_name][
                        model_name
                    ]["failure_stage_CG"] = failure_stage_CG

                    # Failure from NED stage
                    failure_stage_NED = count_miss_NED_per_type.sum() / (
                        count_miss_CG_per_type.sum() + count_miss_NED_per_type.sum()
                    )
                    self.detailed_results_analysis[eval_strat][dataset_name][
                        model_name
                    ]["failure_stage_NED"] = failure_stage_NED

                    # print(f"Number of mentions where the CG step failed for model {model_name} on dataset {dataset_name} with eval strategy {eval_strat} : {count_miss_CG}")
                    # print(f"Number of mentions where the NED step failed for model {model_name} on dataset {dataset_name} with eval strategy {eval_strat} : {count_miss_NED}")
                    # print(f"Number of of correctly linked mentions for model {model_name} on dataset {dataset_name} with eval strategy {eval_strat} : {count_success}")
                    # print(f"Number of mentions with hit_index < k for model {model_name} on dataset {dataset_name} with eval strategy {eval_strat} : {count_hit_k}")
                    # print(f"Accuracy per type for model {model_name} on dataset {name} with eval strategy {eval_strat} : {accuracy_per_type}")
                    # print(f"Recall@k per type for model {model_name} on dataset {name} with eval strategy {eval_strat} : {recall_k_per_type}")
                    # print(" '%' of mentions for which linking failed in CG for model {model_name} on dataset {dataset_name} with eval strategy {eval_strat} : ", failure_stage_CG)
                    # print(" '%' of mentions for which linking failed in NED for model {model_name} on dataset {dataset_name} with eval strategy {eval_strat} : ", failure_stage_NED)

                    # Create a new column candidate generation 'CG' to indicate if the correct cui is among the candidates (1) or not (0)
                    df["CG"] = df[name].apply(lambda x: 1 if x != 1000000 else 0)
                    # Create a new column Named Entity Disambiguation 'NED' to indicate if the prediction was correct (1) or incorrect (0)
                    df["NED"] = df[name].apply(lambda x: 1 if x == 0 else 0)
                    # Explode the list so that an entity is counted for each type it belongs to
                    df_exploded = df.explode("type")

                    # Create contingency tables
                    contingency_table_CG = pd.crosstab(
                        df_exploded["type"], df_exploded["CG"]
                    )
                    contingency_table_NED = pd.crosstab(
                        df_exploded["type"], df_exploded["NED"]
                    )
                    self.detailed_results_analysis[eval_strat][dataset_name][
                        model_name
                    ]["contingency_table_CG"] = contingency_table_CG
                    self.detailed_results_analysis[eval_strat][dataset_name][
                        model_name
                    ]["contingency_table_NED"] = contingency_table_NED

                    # Perform the Chi-square test
                    chi2_NED, p_value_NED, dof_NED, expected_NED = chi2_contingency(
                        contingency_table_NED,
                    )
                    chi2_CG, p_value_CG, dof_CG, expected_CG = chi2_contingency(
                        contingency_table_CG,
                    )
                    self.detailed_results_analysis[eval_strat][dataset_name][
                        model_name
                    ]["ChiSquare_test_NED"] = chi2_contingency(contingency_table_NED)
                    self.detailed_results_analysis[eval_strat][dataset_name][
                        model_name
                    ]["ChiSquare_test_CG"] = chi2_contingency(contingency_table_CG)

                    # # # Print results
                    # # print("Results for success in CG step:")
                    # # print(f"Degree of freedom (dof) = number of different classes : {dof_CG}")
                    # # # observed vs expected frequencies under the assumption of independence
                    # # # If observed < expected, then there is a statistical difference for the class
                    # # print(f"Expected frequencies table :{expected_CG}")
                    # print(f"Chi-square statistic for model {model_name} on dataset {dataset_name} with eval strategy {eval_strat} : {chi2_CG}")
                    # print(f"P-value for model {model_name} on dataset {dataset_name} with eval strategy {eval_strat} : {p_value_CG}")

                    # print('-'*50)
                    # # print("Results for succes in NED step:")
                    # # print(f"Degree of freedom (dof) = number of different classes : {dof_NED}")
                    # # print(f"Expected frequencies table :{expected_NED}") # Expected values for class=0 (failure) and class=1 (success) for the different categories (rows)
                    # print(f"Chi-square statistic for model {model_name} on dataset {dataset_name} with eval strategy {eval_strat} : {chi2_NED}")
                    # print(f"P-value for model {model_name} on dataset {dataset_name} with eval strategy {eval_strat} : {p_value_NED}")

                    precision_k_DK = {}  # David's version of precision@k
                    precision_k = {}
                    map_k = {}

                    # filtered_df = df[df["sapbert_resolve_abbrev"] != 1000000]  # Filter out mentions whose correct answer were not in the candidates
                    for i in range(1, k + 1):
                        precision_k_DK[k] = precision_at_k_DK(
                            df, "sapbert_resolve_abbrev", k
                        )
                        precision_k[i] = precision_at_k(df, name, i)
                        map_k[i] = mean_average_precision_at_k(df, name, i)
                        # precision_k_DK[k] = precision_at_k_DK(filtered_df, "sapbert_resolve_abbrev", k)
                        # precision_k[i] = precision_at_k(filtered_df, name, i)
                        # map_k[i] = mean_average_precision_at_k(filtered_df, name, i)

                    self.detailed_results_analysis[eval_strat][dataset_name][
                        model_name
                    ]["precision@k"] = precision_k
                    self.detailed_results_analysis[eval_strat][dataset_name][
                        model_name
                    ]["map@k"] = map_k
                    self.detailed_results_analysis[eval_strat][dataset_name][
                        model_name
                    ]["precision@k_DK"] = precision_k_DK



from tqdm import tqdm


def _obo_extract_synonyms(data):
    if 'synonym' in data:
        return [syn.split('"')[1]for syn in data['synonym']]
    else:
        return []

def _obo_extract_definition(data):
    if 'def' in data:
        return data['def'].strip('"').split('"')[0]

        
def _obo_term_to_synonyms(graph, filter_prefix=None):
    node_dict = {}
    for curie, data in tqdm(graph.nodes(data=True)):
        if filter_prefix is not None:
            if not curie.startswith(filter_prefix):
                continue
        if 'name' not in data:
            # print(f'Missing name.  CURIE: {curie}, data: {data}')
        # if 'synonym' not in data:
            # print(f'Missing synonym.  CURIE: {curie}, data: {data}')
            synonyms = _obo_extract_synonyms(data)
        else:
            synonyms = [data['name']] + _obo_extract_synonyms(data)
            node_dict[curie] = synonyms

    return node_dict

def _obo_term_to_definitions(graph, filter_prefix=None):
    node_dict = {}
    for curie, data in tqdm(graph.nodes(data=True)):
        if filter_prefix is not None:
            if not curie.startswith(filter_prefix):
                continue
        else:
            definition = _obo_extract_definition(data)
            node_dict[curie] = definition

    return node_dict


import unittest
import numpy as np
import os
import json

from tqdm import tqdm

from bioel.ontology import BiomedicalOntology


class TestOntology(unittest.TestCase):

    def test_medic_loader(self):
        """
        TestCase - 1: Read from the local UMLS Directory and check for data formatting issues in the entities.
        """
        test_cases = [
            {
                "filepath": "/mitchell/entity-linking/kbs/medic.tsv",
                "name": "MEDIC",
                "abbrev": None,
            }
        ]

        for case in tqdm(test_cases):
            ontology = BiomedicalOntology.load_medic(**case)
            print(list(ontology.entities.values())[:5])
            self.check_multiprefix(ontology)
            self.check_unique_cui(ontology)

    def test_json_loader(self):
        """
        TestCase - 1: Read from the Json/Jsonl Ontology Directory and check for data formatting issues in the entities.
        """
        test_cases = [
            {
                "filepath": "https://ai2-s2-scispacy.s3-us-west-2.amazonaws.com/data/kbs/2023-04-23/umls_mesh_2022.jsonl",
                "name": "MESH",
            },
            {
                "filepath": "https://ai2-s2-scispacy.s3-us-west-2.amazonaws.com/data/kbs/2023-04-23/umls_mesh_2022.jsonl",
                "name": "MESH",
            },
            {
                "filepath": "./Cached_files/datasets/umls_mesh_2022.jsonl",
                "name": "MESH",
            }

        ]
        case_index=0
        for case in tqdm(test_cases):
            ontology = BiomedicalOntology.load_json(**case)
            print(list(ontology.entities.values())[:5])
            self.check_multiprefix(ontology)
            self.check_unique_cui(ontology)

            if case_index == 0 :
                self.check_url_caching(case["filepath"])
            if case_index == 1 :
                self.check_unique_url_caching(case["filepath"])
            
            case_index += 1
    
    def test_mesh_loader(self):
        """
        TestCase - 1: Read from the local UMLS Directory and check for data formatting issues in the entities.
        """
        test_cases = [
            {
                "filepath": "/mitchell/entity-linking/2017AA/META/",
                "name": "MESH",
                "abbrev": None,
            }
        ]

        for case in tqdm(test_cases):
            ontology = BiomedicalOntology.load_mesh(**case)
            print(list(ontology.entities.values())[:5])
            self.check_multiprefix(ontology)
            self.check_unique_cui(ontology)
    
    def test_umls_loader(self):
        """
        TestCase - 1: Read from the local UMLS Directory and check for data formatting issues in the entities.
        """
        test_cases = [
            {
                "filepath": "/mitchell/entity-linking/2017AA/META/",
                "name": "UMLS",
                "abbrev": None,
            }
        ]

        for case in tqdm(test_cases):
            ontology = BiomedicalOntology.load_umls(**case)
            print(list(ontology.entities.values())[:5])
            self.check_multiprefix(ontology)
            self.check_unique_cui(ontology)
    
    def test_obo_loader(self):

        test_cases = [
            {
                "filepath": "https://raw.githubusercontent.com/DiseaseOntology/HumanDiseaseOntology/main/src/ontology/doid.obo",
                "name": "disease ontology",
                "prefix_to_keep": None,
                "entity_type": "disease",
                "abbrev": "doid",
            },
            {
                "filepath": "http://purl.obolibrary.org/obo/cl.obo",
                "name": "cell ontology",
                "prefix_to_keep": "CL",
                "entity_type": "cell line & type",
                "abbrev": "cl",
            },
            # {'filepath': 'http://purl.obolibrary.org/obo/uberon.obo',
            # 'name': 'uberon',
            # 'prefix_to_keep': 'UBERON',
            # 'entity_type': 'tissue',
            # 'abbrev': 'uberon'},
            {
                "filepath": "http://purl.obolibrary.org/obo/obi.obo",
                "name": "ontology of biological investigations",
                "prefix_to_keep": "OBI",
                "entity_type": "assay",
                "abbrev": "obi",
            },
            {
                "filepath": "https://ftp.expasy.org/databases/cellosaurus/cellosaurus.obo",
                "name": "cellosaurus",
                "prefix_to_keep": None,
                "entity_type": "cell line & type",
                "abbrev": "cvcl",
            },
            # {'filepath': 'http://purl.obolibrary.org/obo/go.obo',
            # 'name': 'gene ontology',
            # 'prefix_to_keep': None,
            # 'entity_type': 'subcellular',
            # 'abbrev': 'go'},
        ]

        for case in tqdm(test_cases):
            ontology = BiomedicalOntology.load_obo(**case)
            print(list(ontology.entities.values())[:5])
            self.check_multiprefix(ontology)
            self.check_unique_cui(ontology)
    
    def check_multiprefix(self, ontology: BiomedicalOntology):
        """
        Make each entity has a single prefix
        """
        error_raised = False
        errors = []
        for cui, entity in ontology.entities.items():
            if len(entity.cui.split(":")) > 2:
                raise AssertionError(
                    f"Entities should only have one prefix, but the entity {entity.cui} has more than 1"
                )

    def check_unique_cui(self, ontology: BiomedicalOntology):
        """
        Make sure all CUIs in ontology are unique
        """
        cuis = [entity.cui for cui, entity in ontology.entities.items()]
        if len(cuis) != len(set(cuis)):
            raise AssertionError("Ontology contains duplicate CUIs")
    
    def check_url_caching(self, url: str, path = "./Cached_files/datasets"):
        """
        This function is used to be sure that the provided HTTP path to the file is stored in the cache folder correctly 
        """
        # List all files in the directory
        files = os.listdir(path)

        # Filter files ending with .json
        json_files = [file for file in files if file.endswith('.json')]

        # Iterate over each JSON file
        index = 0
        for json_file in json_files:
            file_path = os.path.join(path, json_file)

            with open(file_path, 'r') as f:
                # Load JSON data
                json_data = json.load(f)
                
                # Check if the first key is "url"
                if "url" in json_data:
                    url_value = json_data["url"]
                    if url_value == url :
                        break
            
            if index == len(json_files):
                raise AssertionError(f"Could not find any ontology URL: {url} cached in {path}")
                    

    def check_unique_url_caching(self, url: str, path = "./Cached_files/datasets"):
        """
        This function is used to be sure that the provided HTTP path to the file is stored in the cache folder 
        only once for each unique HTTP path
        """
        # List all files in the directory
        files = os.listdir(path)

        # Filter files ending with .json
        json_files = [file for file in files if file.endswith('.json')]

        num_cached_url = 0
        # Iterate over each JSON file
        for json_file in json_files:
            file_path = os.path.join(path, json_file)

            with open(file_path, 'r') as f:
                # Load JSON data
                json_data = json.load(f)
                
                # Check if the first key is "url"
                if "url" in json_data:
                    url_value = json_data["url"]
                    if url_value == url :
                        num_cached_url += 1

        if num_cached_url >= 2:
            raise AssertionError(f"Ontology URL in {url} has more than 1 cached file in {path}")

if __name__ == "__main__":
    unittest.main()
